{% extends "_base.html" %}
{% block title %}{{ super() }} {{ data['title'] }} {% endblock %}
{% block crumbs %}{{ super() }}
/ <a href="{{ data['collections_path'] }}">{% trans %}Collections{% endtrans %}</a>
/ <a href="../../{{ data['id'] }}">{{ data['title'] }}</a>
/ <a href="../../{{ data['id'] }}/tiles">{% trans %}Tiles{% endtrans %}</a>
/ <a href="../../{{ data['id'] }}/tiles/{{ data['tileset'] }}">Tile Metadata</a>
{% endblock %}
{% block body %}
    <section id="collection">
      <h1>{{ data['title'] }}</h1>
      <p>{{ data['description'] }}</p>
      <p>
        {% for kw in data['keywords'] %}
          <span class="badge text-bg-primary bg-primary">{{ kw }}</span>
        {% endfor %}
      </p>
      <h3>{% trans %}Tiles metadata{% endtrans %} - {{ data['format'] }} {% trans %}format{% endtrans %}</h3>
      <h4>{% trans %}Tileset{% endtrans %} {{ data['tileset'] }}</h4>
      <ul>
      {% for key in data['metadata'].keys() %}
        <li>{{ key }} (<code>{{ data['metadata'][key] }}</code>)</li>
      {% endfor %}
      </ul>
    </section>
{% endblock %}
