{% extends "_base.html" %}
{% block title %}{{ super() }} {{ data['title'] }} {% endblock %}
{% block crumbs %}{{ super() }}
/ <a href="{{ data['collections_path'] }}">{% trans %}Collections{% endtrans %}</a>
/ <a href="../../../collections/{{ data['id'] }}">{{ data['title'] }}</a>
{% endblock %}
{% block body %}
    <section id="collection">
      <h1>{{ data['title'] }}</h1>
      <p>{{ data['description'] }}</p>
      <h3>{% trans %}Coverage range type{% endtrans %}</h3>
      <h4>{% trans %}Fields{% endtrans %}</h4>
      <ul>
	{% for field in data['field'] %}
	<li>{{ field['id'] }}: {{ field['name'] }} ({{ field['definition'] }})</li>
	{% endfor %}
      </ul>
    </section>
{% endblock %}
